import factory


class SQLAlchemyFactoryBase(factory.alchemy.SQLAlchemyModelFactory):
    @classmethod
    async def _create(cls, model_class, **kwargs):
        if not cls._meta.sqlalchemy_session_factory:
            raise AttributeError("sqlalchemy_session_factory not found")

        async_sessionmaker = cls._meta.sqlalchemy_session_factory

        instance = cls.build(**kwargs)

        async with async_sessionmaker() as session:
            session.add(instance)
            await session.flush()
            await session.commit()
            return instance

    @classmethod
    async def create_async(cls, **kwargs):
        return await cls._create(cls._meta.model, **kwargs)

    @classmethod
    async def create_batch_async(cls, size: int, **kwargs):
        return [await cls.create_async(**kwargs) for _ in range(size)]



from brilliance_admin import sqlalchemy
from brilliance_admin.translations import TranslateText as _
from example.sections.models import Terminal


class TerminalAdmin(sqlalchemy.SQLAlchemyAdmin):
    model = Terminal
    title = _('terminals')
    icon = 'mdi-console-network-outline'

    ordering_fields = [
        'id',
    ]
    search_fields = [
        'id',
        'title',
    ]

    table_schema = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Terminal,
        list_display=[
            'id',
            'status',
            'merchant_id',
            'public_id',
            'currency_id',
            'title',
            'is_h2h',
            'imitation_api',
            'test_mode',
            'is_active',
        ],
    )
    table_filters = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Terminal,
        fields=[
            'id',
            'created_at',
            'merchant_id',
            'currency_id',
        ]
    )



from contextlib import asynccontextmanager

from sqlalchemy import event
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

ASYNC_ENGINE = create_async_engine(
    "sqlite+aiosqlite:///:memory:",
    future=True,
)


@event.listens_for(ASYNC_ENGINE.sync_engine, "connect")
def _enable_fk_async(dbapi_connection, _):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


async_sessionmaker_ = async_sessionmaker(
    ASYNC_ENGINE,
    expire_on_commit=False,
    class_=AsyncSession,
)


async def recreate_tables_async():
    from example.sections.models import ModelBase

    async with ASYNC_ENGINE.begin() as conn:
        await conn.run_sync(ModelBase.metadata.drop_all)
        await conn.run_sync(ModelBase.metadata.create_all)


@asynccontextmanager
async def lifespan(app):
    await recreate_tables_async()

    from example.sections import models

    await models.CurrencyFactory.create_batch_async(5)
    await models.MerchantFactory.create_batch_async(10)
    await models.TerminalFactory.create_batch_async(15)
    await models.UserFactory.create_batch_async(27)

    yield



from typing import Any

from brilliance_admin import auth, schema
from brilliance_admin.exceptions import AdminAPIException, APIError, FieldError
from brilliance_admin.integrations.sqlalchemy.fields_schema import SQLAlchemyFieldsSchema
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import get_logger

logger = get_logger()


class SQLAlchemyAdminRetrieveMixin:
    has_retrieve: bool = True

    table_schema: SQLAlchemyFieldsSchema

    async def retrieve(
            self,
            pk: Any,
            user: auth.UserABC,
            language_context: LanguageContext,
    ) -> schema.RetrieveResult:
        if not self.has_delete:
            raise AdminAPIException(APIError(message=_('errors.method_not_allowed')), status_code=500)

        # pylint: disable=import-outside-toplevel
        from sqlalchemy import inspect

        col = inspect(self.model).mapper.columns[self.pk_name]
        python_type = col.type.python_type

        assert self.pk_name
        stmt = self.get_queryset().where(getattr(self.model, self.pk_name) == python_type(pk))

        try:
            async with self.db_async_session() as session:
                record = (await session.execute(stmt)).scalars().first()
                data = await self.table_schema.serialize(
                    record,
                    extra={"record": record, "user": user},
                )

        except FieldError as e:
            logger.exception(
                'SQLAlchemy %s retrieve %s #%s field error: %s',
                type(self).__name__, self.model.__name__, pk, e,
            )
            msg = _('errors.serialize_field_error') % {'error': e.message}
            raise AdminAPIException(APIError(message=msg, code='filters_exception'), status_code=500) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s retrieve %s #%s db error: %s',
                type(self).__name__, self.model.__name__, pk, e,
            )
            raise AdminAPIException(
                APIError(message=_('errors.db_error_retrieve'), code='db_error_retrieve'), status_code=500,
            ) from e

        if record is None:
            msg = _('errors.record_not_found') % {'pk_name': self.pk_name, 'pk': pk}
            raise AdminAPIException(
                APIError(message=msg, code='record_not_found'),
                status_code=400,
            )

        logger.debug(
            '%s model %s #%s retrieved by %s',
            type(self).__name__, self.table_schema.model.__name__, pk, user.username,
            extra={'data': data},
        )
        return schema.RetrieveResult(data=data)



import pytest

from brilliance_admin.translations import LanguageContext, LanguageManager
from example.sqlite import ASYNC_ENGINE, async_sessionmaker_, recreate_tables_async


@pytest.fixture
async def sqlite_sessionmaker():
    await recreate_tables_async()
    try:
        yield async_sessionmaker_
    finally:
        await ASYNC_ENGINE.dispose()


@pytest.fixture
async def language_context():
    language_manager = LanguageManager(
        locales_dir='example/locales',
        languages={
            'ru': 'Russian',
            'en': 'English',
        },
    )
    lc = LanguageContext('ru', language_manager)
    yield lc



from typing import Any, Dict, List

from pydantic import BaseModel, Field

from brilliance_admin.schema import Category
from brilliance_admin.schema.category import GraphInfoSchemaData
from brilliance_admin.schema.table.fields_schema import FieldsSchema
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import SupportsStr


class GraphData(BaseModel):
    search: str | None = None
    filters: Dict[str, Any] = Field(default_factory=dict)


class ChartData(BaseModel):
    data: dict
    options: dict
    width: int | None = None
    height: int = 50
    type: str = 'line'


class GraphsDataResult(BaseModel):
    charts: List[ChartData]


class CategoryGraphs(Category):
    _type_slug: str = 'graphs'

    search_enabled: bool = False
    search_help: SupportsStr | None = None

    table_filters: FieldsSchema | None = None

    def generate_schema(self, user, language_context: LanguageContext) -> GraphInfoSchemaData:
        schema = super().generate_schema(user, language_context)
        graph = GraphInfoSchemaData(
            search_enabled=self.search_enabled,
            search_help=language_context.get_text(self.search_help),
        )

        if self.table_filters:
            graph.table_filters = self.table_filters.generate_schema(user, language_context)

        schema.graph_info = graph
        return schema

    async def get_data(self, data: GraphData, user) -> GraphsDataResult:
        raise NotImplementedError('get_data is not implemented')



import pytest

from brilliance_admin.auth import AuthData
from brilliance_admin.exceptions import AdminAPIException
from brilliance_admin import sqlalchemy
from brilliance_admin.schema.admin_schema import AdminSchemaData
from example.sections.models import User, UserFactory


@pytest.mark.asyncio
async def test_login(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)
    result = await auth.login(data=AuthData(username='123', password='test'))
    assert result.user.username == user.username


@pytest.mark.asyncio
async def test_login_not_admin(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    await UserFactory(username='123', password='test')
    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='123', password='test'))

    assert e.value.get_error().code == 'not_an_admin'


@pytest.mark.asyncio
async def test_login_not_found(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    with pytest.raises(AdminAPIException) as e:
        await auth.login(data=AuthData(username='123', password='test'))

    assert e.value.get_error().code == 'user_not_found'


@pytest.mark.asyncio
async def test_authenticate(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)

    token = auth.get_token(user)
    result_user = await auth.authenticate(headers={'Authorization': f'Token {token}'})
    AdminSchemaData(groups={}, profile=result_user)

    assert result_user.username == user.username


@pytest.mark.asyncio
async def test_authenticate_bad_secret(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test', is_admin=True)

    token = auth.get_token(user)
    result_user = await auth.authenticate(headers={'Authorization': f'Token {token}'})
    AdminSchemaData(groups={}, profile=result_user)
    assert result_user.username == user.username

    auth.secret = 'another'
    with pytest.raises(AdminAPIException) as e:
        await auth.authenticate(headers={'Authorization': f'Token {token}'})

    assert e.value.get_error().code == 'token_error'


@pytest.mark.asyncio
async def test_authenticate_not_admin(sqlite_sessionmaker):
    auth = sqlalchemy.SQLAlchemyJWTAdminAuthentication(
        secret='123',
        db_async_session=sqlite_sessionmaker,
        user_model=User,
    )
    user = await UserFactory(username='123', password='test')
    with pytest.raises(AdminAPIException) as e:
        token = auth.get_token(user)
        await auth.authenticate(headers={'Authorization': f'Token {token}'})

    assert e.value.get_error().code == 'user_not_found'



import pytest
from fastapi.testclient import TestClient

from brilliance_admin.schema.table.admin_action import ActionData
from example.main import app

client = TestClient(app)


@pytest.mark.asyncio
async def test_exception_handle(mocker):
    url = app.url_path_for(
        'table_action',
        group='payments',
        category='payments',
        action='action_with_exception',
    )
    request_data = ActionData()
    response = client.post(url, json=request_data.model_dump(mode='json'))
    assert response.status_code == 500, response.content.decode()
    response_data = {
        'code': 'user_action_error',
        'field_errors': None,
        'message': 'Пример ошибки исключения.',
    }
    assert response.json() == response_data



from typing import Dict

from pydantic import Field
from pydantic.dataclasses import dataclass

from brilliance_admin.utils import DataclassBase, SupportsStr


@dataclass
class FieldError(DataclassBase, Exception):
    message: SupportsStr = None
    code: str | None = None

    def __post_init__(self):
        if not self.message and not self.code:
            msg = 'FieldError must contain message or code'
            raise AttributeError(msg)


@dataclass
class APIError(DataclassBase):
    message: SupportsStr | None = None
    code: str | None = None
    field_errors: Dict[str, FieldError] | None = None


@dataclass
class AdminAPIException(DataclassBase, Exception):
    error: APIError = Field(default_factory=APIError)
    status_code: int = 400
    error_code: str | None = None

    def __str__(self):
        return str(self.error)

    def get_error(self) -> APIError:
        return self.error



import abc
from typing import Dict, List

from pydantic.dataclasses import dataclass

from brilliance_admin.auth import UserABC
from brilliance_admin.schema.category import Category, CategorySchemaData
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import DataclassBase, SupportsStr, get_logger

logger = get_logger()


@dataclass
class GroupSchemaData(DataclassBase):
    title: str | None
    description: str | None
    icon: str | None
    categories: Dict[str, CategorySchemaData]


@dataclass
class Group(abc.ABC):
    categories: List[Category]
    slug: str
    title: SupportsStr | None = None
    description: SupportsStr | None = None

    # https://pictogrammers.com/library/mdi/
    icon: str | None = None

    def __post_init__(self):
        for category in self.categories:
            if not issubclass(category.__class__, Category):
                raise TypeError(f'Category "{category}" is not instance of Category subclass')

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> GroupSchemaData:
        result = GroupSchemaData(
            title=language_context.get_text(self.title) or self.slug,
            description=language_context.get_text(self.description),
            icon=self.icon,
            categories={},
        )
        if not self.categories:
            logger.warning('Group "%s" %s.categories is empty!', self.slug, type(self).__name__)

        for category in self.categories:

            if not category.slug:
                msg = f'Category {type(category).__name__}.slug is empty'
                raise AttributeError(msg)

            if category.slug in result.categories:
                exists = result.categories[category.slug]
                msg = f'Category {type(category).__name__}.slug "{self.slug}" already registered by "{exists.title}"'
                raise KeyError(msg)

            result.categories[category.slug] = category.generate_schema(user, language_context)

        return result

    def get_category(self, category_slug: str) -> Category | None:
        for category in self.categories:
            if category.slug == category_slug:
                return category

        return None



import asyncio
from typing import Any, ClassVar, Dict, List

from pydantic_core import core_schema

from brilliance_admin.auth import UserABC
from brilliance_admin.exceptions import AdminAPIException, APIError, FieldError
from brilliance_admin.schema.category import FieldSchemaData, FieldsSchemaData
from brilliance_admin.schema.table.fields.base import TableField
from brilliance_admin.schema.table.fields.function_field import FunctionField
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import DeserializeAction

NOT_FUND_EXCEPTION = '''Field slug "{field_slug}" not found inside generated fields inside {class_name}
Available options: {available_fields}
'''


class DeserializeError(Exception):
    pass


class FieldsSchema:
    # Список полей
    fields: List[str] | None = None

    # Список колонок, которые будут отображаться в таблице
    list_display: List[str] | None = None

    # Для передачи параметра read_only = True внутрь поля
    readonly_fields: ClassVar[List | None] = None

    # Generated fields
    _generated_fields: dict = None

    def __init__(self, *args, table_schema=None, list_display=None, readonly_fields=None, fields=None, **kwargs):
        if fields:
            self.fields = fields

        if self.fields and not isinstance(self.fields, list):
            msg = f'{type(self).__name__}.fields must be a list instance; found: {self.fields}'
            raise AttributeError(msg)

        if table_schema:
            self.table_schema = table_schema

        if list_display:
            self.list_display = list_display

        if readonly_fields:
            self.readonly_fields = readonly_fields

        generated_fields = self.generate_fields(kwargs)

        available_fields = list(generated_fields.keys())
        if self.fields is None:
            self.fields = available_fields

        self._generated_fields = {}
        for field_slug in self.fields:
            if not isinstance(field_slug, str):
                msg = f'{type(self).__name__} field "{field_slug}" must be string'
                raise AttributeError(msg)

            if field_slug not in generated_fields:
                msg = NOT_FUND_EXCEPTION.format(
                    field_slug=field_slug,
                    available_fields=available_fields,
                    class_name=type(self).__name__,
                )
                raise AttributeError(msg)

            self._generated_fields[field_slug] = generated_fields[field_slug]

        self.validate_fields(*args, **kwargs)

    def validate_fields(self, *args, **kwargs):
        if not self.fields:
            msg = f'Schema {type(self).__name__}.fields is empty'
            raise AttributeError(msg)

        # Check for fields not listed in self.fields
        for attribute_name in dir(self):
            if '__' in attribute_name:
                continue

            attribute = getattr(self, attribute_name)
            if issubclass(attribute.__class__, TableField) and attribute_name not in self.fields:
                msg = f'Schema {type(self).__name__} attribute "{attribute_name}" {type(attribute).__name__} presented, but not listed inside fields list: {self.fields}'
                raise AttributeError(msg)

        if self.readonly_fields:
            for field_slug in self.readonly_fields:
                field = self.get_field(field_slug)
                if not field:
                    msg = f'{type(self).__name__} field "{field_slug}" from readonly_fields is not found inside fields; available options: {self.fields}'
                    raise AttributeError(msg)

                field.read_only = True

        # Fill list_display
        if self.list_display is None:
            self.list_display = self.fields

        for field_slug in self.list_display:
            if field_slug not in self.fields:
                msg = f'Field "{field_slug}" inside {type(self).__name__}.list_display, but not presented as field; available options: {self.fields}'
                raise AttributeError(msg)

    def generate_fields(self, kwargs) -> dict:
        generated_fields = {}

        # Fields from kwargs
        for k, v in kwargs.items():
            if issubclass(type(v), TableField) and not hasattr(self, k):
                generated_fields[k] = v

        # Autogenerate fields from instance attributes
        for attribute_name in dir(self):
            if '__' in attribute_name:
                continue

            attribute = getattr(self, attribute_name)
            if issubclass(type(attribute), TableField):
                generated_fields[attribute_name] = attribute

        # Generation FunctionField
        for attribute_name in dir(self):
            if '__' in attribute_name:
                continue

            attribute = getattr(self, attribute_name)
            if getattr(attribute, '__function_field__', False):
                field = FunctionField(fn=attribute, **attribute.__kwargs__)
                field.read_only = True
                generated_fields[attribute_name] = field

        return generated_fields

    def get_field(self, field_slug) -> TableField | None:
        return self.get_fields().get(field_slug)

    def get_fields(self) -> Dict[str, TableField]:
        return self._generated_fields

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> FieldsSchemaData:
        fields_schema = FieldsSchemaData(
            list_display=self.list_display,
        )

        context = {'language_context': language_context}
        for field_slug, field in self.get_fields().items():
            field_schema: FieldSchemaData = field.generate_schema(user, field_slug, language_context)
            fields_schema.fields[field_slug] = field_schema.to_dict(keep_none=False, context=context)

        return fields_schema

    async def serialize(self, data: Any, extra: dict) -> dict:
        result = {}
        for field_slug, field in self.get_fields().items():
            value = data.get(field_slug)
            result[field_slug] = await field.serialize(value, extra)
        return result

    async def deserialize(self, data: dict, action: DeserializeAction, extra) -> dict:
        result = {}
        errors = {}
        for field_slug, field in self.get_fields().items():

            if field.read_only and action != DeserializeAction.FILTERS:
                continue

            # Skip update if fields is not presented in data
            if action in [DeserializeAction.UPDATE, DeserializeAction.FILTERS] and field_slug not in data:
                continue

            value = data.get(field_slug)
            try:
                deserialized_value = await field.deserialize(value, action, extra)

                validate_method = getattr(self, f'validate_{field_slug}', None)
                if callable(validate_method):
                    if not asyncio.iscoroutinefunction(validate_method):
                        msg = f'Validate method {type(self).__name__}.{field_slug} must be async'
                        raise AttributeError(msg)
                    deserialized_value = await validate_method(value)

                field.set_deserialized_value(result, field_slug, deserialized_value, action, extra)
            except FieldError as e:
                errors[field_slug] = e

        if errors:
            raise AdminAPIException(
                APIError(
                    message='Validation error',
                    code='validation_error',
                    field_errors=errors,
                ),
                status_code=400,
            )
        return result

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: Any) -> core_schema.CoreSchema:
        def validate(v: Any) -> "FieldsSchema":
            if isinstance(v, cls):
                return v
            raise TypeError(f"Expected {cls.__name__} instance")

        return core_schema.no_info_plain_validator_function(
            validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: repr(v),
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )



import abc
import datetime
from enum import Enum
from typing import Any, ClassVar

from pydantic.dataclasses import dataclass

from brilliance_admin.exceptions import FieldError
from brilliance_admin.schema.category import FieldSchemaData
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import DeserializeAction, SupportsStr, humanize_field_name


@dataclass
class TableField(abc.ABC, FieldSchemaData):
    _type: ClassVar[str]

    label: SupportsStr | None = None
    help_text: SupportsStr | None = None

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = FieldSchemaData(
            type=self._type,
            label=language_context.get_text(self.label) or humanize_field_name(field_slug),
            help_text=language_context.get_text(self.help_text),
            header=self.header,
            read_only=self.read_only,
            default=self.default,
            required=self.required,
        )

        return schema

    async def serialize(self, value, extra: dict, *args, **kwargs) -> Any:
        return value

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        if self.required and value is None:
            raise FieldError('Field is required', 'field_required')

        return value

    async def autocomplete(self, model, data, user):
        raise NotImplementedError('autocomplete is not implemented')

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def set_deserialized_value(self, result: dict, field_slug, deserialized_value, action, extra):
        result[field_slug] = deserialized_value


@dataclass
class IntegerField(TableField):
    _type = 'integer'

    choices: Any | None = None

    max_value: int | None = None
    min_value: int | None = None

    inputmode: str | None = None
    precision: int | None = None
    scale: int | None = None

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        if self.max_value is not None:
            schema.max_value = self.max_value

        if self.min_value is not None:
            schema.min_value = self.min_value

        schema.inputmode = self.inputmode
        schema.precision = self.precision
        schema.scale = self.scale

        return schema

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        value = await super().deserialize(value, action, extra, *args, **kwargs)
        if value and not isinstance(value, int):
            raise FieldError(f'bad int type: {type(value)}')

        return value


@dataclass
class StringField(TableField):
    _type = 'string'

    multilined: bool | None = None
    ckeditor: bool | None = None
    tinymce: bool | None = None

    min_length: int | None = None
    max_length: int | None = None

    choices: Any | None = None

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        schema.multilined = self.multilined
        schema.ckeditor = self.ckeditor
        schema.tinymce = self.tinymce

        if self.min_length is not None:
            schema.min_length = self.min_length

        if self.max_length is not None:
            schema.max_length = self.max_length

        return schema

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        value = await super().deserialize(value, action, extra, *args, **kwargs)
        if value and not isinstance(value, str):
            raise FieldError(f'bad string type: {type(value)}')

        return value


@dataclass
class BooleanField(TableField):
    _type = 'boolean'


@dataclass
class DateTimeField(TableField):
    _type = 'datetime'

    format: str = '%Y-%m-%dT%H:%M:%S'
    range: bool | None = None
    include_date: bool | None = True
    include_time: bool | None = True

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        schema.range = self.range
        schema.include_date = self.include_date
        schema.include_time = self.include_time

        return schema

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        value = await super().deserialize(value, action, extra, *args, **kwargs)

        if not value:
            return

        if value and not isinstance(value, (str, dict)):
            raise FieldError(f'bad datetime type: {type(value)}')

        if isinstance(value, str):
            return datetime.datetime.strptime(value, self.format)

        if isinstance(value, dict):
            if not value.get('from') or not value.get('to'):
                msg = f'{type(self).__name__} value must be dict with from,to values: {value}'
                raise FieldError(msg)

            return {
                'from': datetime.datetime.strptime(value.get('from'), self.format),
                'to': datetime.datetime.strptime(value.get('to'), self.format),
            }

        raise NotImplementedError(f'Value "{value}" is not supporetd for datetime')


@dataclass
class JSONField(TableField):
    _type = 'json'


@dataclass
class ArrayField(TableField):
    _type = 'array'

    array_type: str | None

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        schema.array_type = self.array_type

        return schema

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        value = await super().deserialize(value, action, extra, *args, **kwargs)

        if value and not isinstance(value, list):
            raise FieldError(f'bad array type: {type(value)}')

        return value


@dataclass
class FileField(TableField):
    _type = 'file'


@dataclass
class ImageField(TableField):
    _type = 'image'

    preview_max_height: int = 100
    preview_max_width: int = 100

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        if self.preview_max_height is not None:
            schema.preview_max_height = self.preview_max_height

        if self.preview_max_width is not None:
            schema.preview_max_width = self.preview_max_width

        return schema

    async def serialize(self, value, extra: dict, *args, **kwargs) -> Any:
        return {'url': value}


@dataclass
class ChoiceField(TableField):
    _type = 'choice'

    # Tag color available:
    # https://vuetifyjs.com/en/styles/colors/#classes
    choices: Any | None = None

    # https://vuetifyjs.com/en/components/chips/#color-and-variants
    variant: str = 'elevated'
    size: str = 'default'

    def __post_init__(self):
        self.choices = self.generate_choices()

    def generate_choices(self):
        if not self.choices:
            return None

        if issubclass(self.choices, Enum):
            return [
                {'value': c.value, 'title': c.label, 'tag_color': getattr(c, 'tag_color', None)}
                for c in self.choices
            ]

        msg = f'Field choices is not suppored: {self.choices}'
        raise NotImplementedError(msg)

    def find_choice(self, value):
        if not self.choices:
            return None

        return next((c for c in self.choices if c.get('value') == value), None)

    def generate_schema(self, user, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)

        schema.choices = self.choices

        schema.size = self.size
        schema.variant = self.variant

        return schema

    async def serialize(self, value, extra: dict, *args, **kwargs) -> Any:
        if not value:
            return

        choice = self.find_choice(value)
        return {
            'value': value,
            'title': choice.get('title') or value if choice else value.capitalize(),
        }



from brilliance_admin import schema
from brilliance_admin.schema.graphs.category_graphs import ChartData, GraphData, GraphsDataResult
from brilliance_admin.translations import TranslateText as _


class GraphsFiltersSchema(schema.FieldsSchema):
    id = schema.IntegerField(label='ID')
    created_at = schema.DateTimeField(label=_('created_at'))

    _fields = [
        'id',
        'created_at',
    ]


class GraphsExample(schema.CategoryGraphs):
    slug = 'graphs-example'
    title = _('graphs_example')
    icon = 'mdi-chart-bar-stacked'

    table_filters = GraphsFiltersSchema()

    async def get_data(self, data: GraphData, user) -> GraphsDataResult:
        return GraphsDataResult(
            charts=[
                ChartData(
                    type='line',
                    data={
                        'labels': ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul"],
                        'datasets': [
                            {
                                'label': "Dataset #1",
                                'backgroundColor': "rgba(255,99,132,0.2)",
                                'borderColor': "rgba(255,99,132,1)",
                                'borderWidth': 2,
                                'hoverBackgroundColor': "rgba(255,99,132,0.4)",
                                'hoverBorderColor': "rgba(255,99,132,1)",
                                'data': [65, 59, 20, 81, 56, 55, 40],
                            },
                            {
                                'label': "Dataset #2",
                                'backgroundColor': "rgba(233, 150, 122,0.2)",
                                'borderColor': "rgba(233, 150, 122,1)",
                                'borderWidth': 2,
                                'hoverBackgroundColor': "rgba(233, 150, 122,0.4)",
                                'hoverBorderColor': "rgba(233, 150, 122,1)",
                                'data': [30, 35, 29, 15, 3, 10, 22],
                            },
                        ],
                    },
                    options={
                        'responsive': True,
                        'plugins': {
                            'legend': {
                                'position': 'top',
                            },
                            'title': {'display': True, 'text': 'Chart.js Line Chart'},
                        },
                    },
                ),
                ChartData(
                    type='line',
                    data={
                        'labels': [
                            "Jun 2016",
                            "Jul 2016",
                            "Aug 2016",
                            "Sep 2016",
                            "Oct 2016",
                            "Nov 2016",
                            "Dec 2016",
                            "Jan 2017",
                            "Feb 2017",
                            "Mar 2017",
                            "Apr 2017",
                            "May 2017",
                        ],
                        'datasets': [
                            {
                                'label': "Rainfall",
                                'backgroundColor': 'lightblue',
                                'borderColor': 'royalblue',
                                'data': [26.4, 39.8, 66.8, 66.4, 40.6, 55.2, 77.4, 69.8, 57.8, 76, 110.8, 142.6],
                            }
                        ],
                    },
                    options={
                        'layout': {
                            'padding': 10,
                        },
                        'legend': {
                            'position': 'bottom',
                        },
                        'title': {'display': True, 'text': 'Precipitation in Toronto'},
                        'scales': {
                            'yAxes': [{'scaleLabel': {'display': True, 'labelString': 'Precipitation in mm'}}],
                            'xAxes': [{'scaleLabel': {'display': True, 'labelString': 'Month of the Year'}}],
                        },
                    },
                ),
                ChartData(
                    type='bar',
                    data={
                        'labels': ['Red', 'Blue', 'Yellow', 'Green', 'Purple', 'Orange'],
                        'datasets': [
                            {
                                'label': 'Vote Count',
                                'data': [12, 19, 3, 5, 2, 3],
                                'backgroundColor': [
                                    'rgba(255, 99, 132, 0.2)',
                                    'rgba(54, 162, 235, 0.2)',
                                    'rgba(255, 205, 86, 0.2)',
                                    'rgba(75, 192, 192, 0.2)',
                                    'rgba(153, 102, 255, 0.2)',
                                    'rgba(255, 159, 64, 0.2)',
                                ],
                                'borderColor': [
                                    'rgb(255, 99, 132)',
                                    'rgb(54, 162, 235)',
                                    'rgb(255, 205, 86)',
                                    'rgb(75, 192, 192)',
                                    'rgb(153, 102, 255)',
                                    'rgb(255, 159, 64)',
                                ],
                                'borderWidth': 1,
                            }
                        ],
                    },
                    height=50,
                    options={
                        'scales': {'x': {'beginAtZero': True, 'ticks': {'color': '#333'}}, 'y': {'ticks': {'color': '#333'}}},
                        'animation': {'duration': 1500, 'easing': 'easeInOutQuad'},
                    },
                ),
            ]
        )



# pylint: disable=wildcard-import, unused-wildcard-import, unused-import
# flake8: noqa: F405
from .admin_schema import AdminSchema, AdminSchemaData
from .category import Category
from .graphs import *
from .group import Group
from .table import *



from brilliance_admin import schema, sqlalchemy
from brilliance_admin.translations import TranslateText as _
from example.sections.models import User


class UserAdmin(sqlalchemy.SQLAlchemyAdmin):
    model = User
    title = _('users')
    icon = 'mdi-account-details'

    ordering_fields = [
        'id',
    ]
    search_fields = [
        'username',
    ]

    table_schema = sqlalchemy.SQLAlchemyFieldsSchema(
        model=User,
    )
    table_filters = sqlalchemy.SQLAlchemyFieldsSchema(
        model=User,
        created_at=schema.DateTimeField(range=True),
        last_login=schema.DateTimeField(range=True),
    )



import random
import uuid
from datetime import datetime
from enum import Enum

import factory
from sqlalchemy import BigInteger, Boolean, DateTime, ForeignKey, Integer, SmallInteger, String, func
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship
from sqlalchemy.sql import expression

from brilliance_admin.translations import TranslateText as _
from example.sqlite import async_sessionmaker_
from example.utils import SQLAlchemyFactoryBase


class ModelBase(DeclarativeBase):
    pass


class BaseIDModel(ModelBase):
    __abstract__ = True
    id: Mapped[int] = mapped_column(
        BigInteger().with_variant(Integer(), "sqlite"),
        primary_key=True,
        autoincrement=True,
    )


class User(BaseIDModel):
    __tablename__ = "user"

    username: Mapped[str] = mapped_column(String(50), unique=True, index=True, nullable=False)
    email: Mapped[str] = mapped_column(String(120), unique=True, index=True, nullable=False)
    password: Mapped[str] = mapped_column(String(255), nullable=True)

    is_staff: Mapped[bool] = mapped_column(default=False, nullable=False)
    is_admin: Mapped[bool] = mapped_column(default=False, nullable=False)
    is_active: Mapped[bool] = mapped_column(nullable=False, server_default=expression.true())

    last_login: Mapped[datetime] = mapped_column(DateTime(timezone=True), nullable=True)

    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), server_default=func.now(), nullable=False)  # pylint: disable=not-callable

    # group_associations = relationship("UserGroup", back_populates="user")

    def __repr__(self):
        return f"<User(id={self.id}, username='{self.username}', email='{self.email}')>"


class UserFactory(SQLAlchemyFactoryBase):
    class Meta:
        model = User
        sqlalchemy_session_factory = async_sessionmaker_
        sqlalchemy_session_persistence = "commit"

    username = factory.Sequence(lambda n: f"user_{n}")
    email = factory.Faker("email")


class Currency(BaseIDModel):
    __tablename__ = "currency"

    title: Mapped[str] = mapped_column(String(255), nullable=False)
    num_code: Mapped[int] = mapped_column(SmallInteger, unique=True, index=True, nullable=False)
    char_code: Mapped[str] = mapped_column(String(10), unique=True, index=True, nullable=False)

    depth: Mapped[int] = mapped_column(SmallInteger, index=True, nullable=False, default=2)

    terminals: Mapped[list["Terminal"]] = relationship(back_populates="currency")  # noqa F821

    def __repr__(self):
        return f"<Currency(id={self.id}, num_code='{self.num_code}', char_code='{self.char_code}')>"

    def __str__(self):
        return self.title


class CurrencyFactory(SQLAlchemyFactoryBase):
    class Meta:
        model = Currency
        sqlalchemy_session_factory = async_sessionmaker_
        sqlalchemy_session_persistence = "commit"

    title = factory.Faker("currency_name")
    num_code = factory.Sequence(lambda n: 100 + n)
    char_code = factory.Sequence(lambda n: f"C{n:03d}")
    depth = factory.Faker("random_element", elements=[2, 3])


class Merchant(BaseIDModel):
    __tablename__ = "merchant"

    user_id: Mapped[int] = mapped_column(BigInteger, nullable=False, index=True)
    title: Mapped[str] = mapped_column(String(255), nullable=False)
    description: Mapped[str] = mapped_column(String(255), nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True), server_default=func.now(), nullable=False)  # pylint: disable=not-callable
    terminals: Mapped[list["Terminal"]] = relationship(back_populates="merchant")

    def __repr__(self):
        return f"<Merchant(id={self.id}, title='{self.title}')>"

    def __str__(self):
        return self.title


class MerchantFactory(SQLAlchemyFactoryBase):
    class Meta:
        model = Merchant
        sqlalchemy_session_factory = async_sessionmaker_
        sqlalchemy_session_persistence = "commit"

    user_id = factory.Faker("random_int", min=1, max=10_000)
    title = factory.Faker("word")
    description = factory.Faker("sentence", nb_words=6)
    created_at = factory.LazyFunction(datetime.utcnow)

    def __str__(self):
        return self.title


class Fee(BaseIDModel):
    __tablename__ = "fee"

    title: Mapped[str] = mapped_column(String(255), nullable=False)
    terminals: Mapped[list["Terminal"]] = relationship(back_populates="fee")

    def __repr__(self):
        return f"<Fee(id={self.id}, title='{self.title}')>"

    def __str__(self):
        return self.title


class FeeFactory(SQLAlchemyFactoryBase):
    class Meta:
        model = Fee
        sqlalchemy_session_factory = async_sessionmaker_
        sqlalchemy_session_persistence = "commit"

    title = factory.Faker("word")

    def __str__(self):
        return self.title


class TerminalStatuses(Enum):
    PROCESS = 'process'
    SUCCESS = 'success'
    ERROR = 'error'

    @property
    def label(self):
        labels = {
            self.PROCESS: _('statuses.process'),
            self.SUCCESS: _('statuses.success'),
            self.ERROR: _('statuses.error'),
        }
        return labels[self]

    @property
    def tag_color(self):
        colors = {
            self.PROCESS: 'grey-lighten-1',
            self.SUCCESS: 'green-darken-1',
            self.ERROR: 'red-lighten-2',
        }
        return colors[self]


class Terminal(BaseIDModel):
    __tablename__ = "terminal"

    title: Mapped[str] = mapped_column(String(255), nullable=False)
    description: Mapped[str] = mapped_column(
        String(255),
        nullable=False,
        info={"label": _("description")},
    )
    secret_key: Mapped[str] = mapped_column(
        String(255),
        nullable=False,
        default=lambda: str(uuid.uuid4()),
    )

    status: Mapped[str] = mapped_column(
        String(255),
        nullable=False,
        info={"choices": TerminalStatuses},
    )

    return_url: Mapped[str] = mapped_column(String(500), nullable=True)
    callback_url: Mapped[str] = mapped_column(String(500), nullable=True)

    merchant_id: Mapped[int] = mapped_column(ForeignKey("merchant.id"), index=True)
    merchant: Mapped["Merchant"] = relationship(back_populates="terminals")

    currency_id: Mapped[int] = mapped_column(ForeignKey("currency.id"), index=True)
    currency: Mapped["Currency"] = relationship(back_populates="terminals")  # noqa F821

    fee_id: Mapped[int] = mapped_column(ForeignKey("fee.id"), nullable=True)
    fee: Mapped["Fee"] = relationship(back_populates="terminals")

    is_h2h: Mapped[bool] = mapped_column(Boolean, nullable=False, server_default=expression.true())
    is_active: Mapped[bool] = mapped_column(Boolean, nullable=False, server_default=expression.true())

    public_id: Mapped[str] = mapped_column(
        String(255),
        nullable=False,
        default=lambda: str(uuid.uuid4()),
        index=True,
    )
    imitation_api: Mapped[str] = mapped_column(String(50), nullable=True)
    test_mode: Mapped[bool] = mapped_column(nullable=False, default=False)

    registered_delay: Mapped[int | None] = mapped_column(Integer, nullable=True)

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )  # DateTime used once

    # whitelist_ips: Mapped[list[str]] = mapped_column(ARRAY(String), default=list)

    def __repr__(self):
        return f"<Terminal #{self.id} title={self.title}>"

    def __str__(self):
        return self.title


class TerminalFactory(SQLAlchemyFactoryBase):
    class Meta:
        model = Terminal
        sqlalchemy_session_factory = async_sessionmaker_
        sqlalchemy_session_persistence = "commit"

    title = factory.Faker("company")
    description = factory.Faker("sentence", nb_words=6)

    status = factory.LazyFunction(lambda: random.choice(list(TerminalStatuses)).value)

    secret_key = factory.LazyFunction(lambda: str(uuid.uuid4()))
    public_id = factory.LazyFunction(lambda: str(uuid.uuid4()))

    merchant = factory.SubFactory(MerchantFactory)
    currency = factory.SubFactory(CurrencyFactory)

    is_h2h = factory.Faker("boolean")
    is_active = factory.Faker("boolean")

    imitation_api = factory.Faker("random_element", elements=[None, "sandbox", "mock"])
    test_mode = factory.Faker("boolean")

    registered_delay = factory.Faker(
        "random_element",
        elements=[None, 5, 10, 30, 60],
    )



import logging
import re
from pathlib import Path
from typing import Any, Dict, Protocol

import yaml
from pydantic import TypeAdapter
from pydantic_core import core_schema


class SupportsStr(Protocol):
    def __str__(self) -> str: ...

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        def validate(v):
            if not hasattr(v, '__str__'):
                raise TypeError(f'value must implement __str__, got {type(v)}')
            return v

        return core_schema.no_info_plain_validator_function(validate)


def get_logger():
    try:
        # pylint: disable=import-outside-toplevel
        import structlog
        return structlog.get_logger('brilliance_admin')
    except ImportError:
        return logging.getLogger('brilliance_admin')


class DeserializeAction:
    CREATE = 0
    UPDATE = 1
    TABLE_ACTION = 2
    FILTERS = 3


class DataclassBase:
    def model_dump(self, *args, **kwargs) -> dict:
        adapter = TypeAdapter(type(self))
        return adapter.dump_python(self, *args, **kwargs)

    def to_dict(self, *args, keep_none=True, **kwargs) -> dict:
        data = self.model_dump(*args, **kwargs)
        return {
            k: v for k, v in data.items()
            if v is not None and not keep_none
        }


def humanize_field_name(name: str) -> str:
    # Convert snake_case / kebab-case / mixed tokens to Title Case with acronyms preserved
    s = name.replace("-", "_")
    parts = [p for p in s.split("_") if p]

    def cap(token: str) -> str:
        # Keep common acronyms uppercase
        if token.lower() in {"id", "ip", "url", "api", "http", "https", "h2h"}:
            return token.upper()
        # If token contains digits, capitalize first letter only (e.g. "h2h" -> "H2h")
        if re.search(r"\d", token):
            return token[:1].upper() + token[1:].lower()
        return token[:1].upper() + token[1:].lower()

    return " ".join(cap(p) for p in parts)


def iter_locale_files(directory) -> list[Path]:
    if isinstance(directory, str):
        directory = Path(directory)

    if not directory.exists():
        raise FileNotFoundError(directory)
    if not directory.is_dir():
        raise NotADirectoryError(directory)

    for path in directory.rglob('*'):
        if not path.is_file():
            continue

        yield path


def merge_dict_data(base: dict, extra: dict) -> dict:
    if not isinstance(base, dict):
        raise TypeError('base must be dict')
    if not isinstance(extra, dict):
        raise TypeError('extra must be dict')

    result = base.copy()
    for key, value in extra.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = merge_dict_data(result[key], value)
        else:
            result[key] = value
    return result


class YamlI18n:
    data: Dict[str, Any] = {}

    def load_folder(self, path):
        for file_path in iter_locale_files(path):
            with file_path.open(encoding='utf-8') as f:
                data = yaml.safe_load(f)

            if not isinstance(data, dict):
                msg = f'YAML root must be dict: {file_path}, got {type(data)}'
                raise TypeError(msg)

            language = file_path.stem
            if language not in self.data:
                self.data[language] = {}

            if not isinstance(self.data[language], dict):
                raise TypeError(f'language root must be dict: {language}')

            self.data[language] = merge_dict_data(self.data[language], data)

    def get_text(self, slug, language, default_language):
        if not isinstance(slug, str):
            raise TypeError(f'slug must be str, got {type(slug)}')

        if not isinstance(language, str):
            raise TypeError(f'language must be str, got {type(language)}')

        if not isinstance(default_language, str):
            raise TypeError(f'default_language must be str, got {type(default_language)}')

        if not self.data:
            raise ValueError('i18n data is empty')

        for lang in (language, default_language):
            if lang not in self.data:
                continue

            node = self.data[lang]

            for part in slug.split('.'):
                if not isinstance(node, dict):
                    node = None
                    break
                if part not in node:
                    node = None
                    break
                node = node[part]

            if isinstance(node, str):
                return node

        return slug



import asyncio
import datetime
import uuid
from typing import Any

from faker import Faker

from brilliance_admin import auth, schema
from brilliance_admin.exceptions import FieldError
from brilliance_admin.schema.table.admin_action import ActionData, ActionMessage, ActionResult, admin_action
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import get_logger

logger = get_logger()


class PaymentFiltersSchema(schema.FieldsSchema):
    id = schema.IntegerField(label='ID')
    created_at = schema.DateTimeField(label=_('created_at'), range=True)

    fields = [
        'id',
        'created_at',
    ]


class PaymentFieldsSchema(schema.FieldsSchema):
    id = schema.IntegerField(label='ID', read_only=True)
    amount = schema.IntegerField(label=_('amount'))
    endpoint = schema.StringField(label=_('endpoint'))
    description = schema.StringField(label=_('description'))
    other_field = schema.StringField(read_only=True)
    whitelist_ips = schema.ArrayField(label=_('whitelist_ips'))
    # image = schema.ImageField(label=_('image'))
    created_at = schema.DateTimeField(label=_('created_at'), read_only=True)

    list_display = [
        'id',
        'amount',
        'endpoint',
        'description',
        'created_at',
        'get_provider_registry',
        'get_provider_registry_info',
    ]

    @schema.function_field(label=_('registry_checked'), type=schema.BooleanField)
    async def get_provider_registry(self, record, user, **kwargs):
        return True

    @schema.function_field(label=_('registry_info_checked'), type=schema.BooleanField)
    async def get_provider_registry_info(self, record, user, **kwargs):
        return False


class CreatePaymentSchema(schema.FieldsSchema):
    amount = schema.IntegerField(label=_('amount'))
    is_throw_error = schema.BooleanField(label=_('is_throw_error'))

    async def validate_is_throw_error(self, value):
        if value:
            raise FieldError(_('throw_error'))
        return value


class PaymentsAdmin(schema.CategoryTable):
    has_update = False
    has_create = False

    slug = 'payments'
    title = _('payments')
    description = _('payments_description')
    icon = 'mdi-credit-card-outline'

    search_enabled = True
    search_help = _('payments_search_fields')

    table_filters = PaymentFiltersSchema()
    table_schema = PaymentFieldsSchema(readonly_fields=['amount'])
    pk_name = 'id'
    ordering_fields = [
        'id',
    ]

    @admin_action(
        title=_('create_payment'),
        description=_('create_payment_description'),
        form_schema=CreatePaymentSchema(),
        allow_empty_selection=True,
    )
    async def create_payment(self, action_data: ActionData):
        await asyncio.sleep(1)
        fake = Faker()
        msg = _('payment_create_result') % {
            'gateway_id': str(uuid.uuid4()),
            'desctiption': fake.sentence(nb_words=100),
            'redirect_url': 'https://www.google.com',
        }
        return ActionResult(persistent_message=msg)

    @admin_action(
        title=_('delete'),
        confirmation_text=_('delete_confirmation_text'),
        base_color='red-lighten-2',
        variant='outlined',
    )
    async def delete(self, action_data: ActionData):
        await asyncio.sleep(1)
        return ActionResult(message=ActionMessage(_('deleted_successfully')))

    @admin_action(title=_('action_with_exception'), allow_empty_selection=True)
    async def action_with_exception(self, action_data: ActionData):
        await asyncio.sleep(0.5)
        raise Exception(_('exception_example'))

    def _get_data(self, pk):
        fake = Faker()
        Faker.seed(pk)

        statuses = list(STATUS_COLORS.keys())
        status = statuses[fake.pyint(min_value=0, max_value=len(statuses) - 1)]
        return {
            'id': pk,
            'amount': 10 * fake.pyint(min_value=0, max_value=100),
            'status': status,
            'endpoint': fake.word(),
            'whitelist_ips': ['localhost', '0.0.0.0'],
            'description': fake.sentence(nb_words=5),
            'other_field': fake.word(),
            'image': f'https://picsum.photos/id/{5039-pk+1}/200/300',
            'created_at': datetime.datetime(2025, 6, 16, 9, 45, 29) - datetime.timedelta(hours=pk, minutes=pk),
        }

    # pylint: disable=too-many-arguments
    async def get_list(
        self,
        list_data: schema.ListData,
        user: auth.UserABC,
        language_context: LanguageContext,
    ) -> schema.TableListResult:
        await asyncio.sleep(0.2)

        data = []
        total_count = 5039

        for i in range(0, list_data.limit):
            pk = total_count - ((list_data.page - 1) * list_data.limit + i)
            if pk < 0:
                continue

            line_data = self._get_data(pk)
            line = await self.table_schema.serialize(line_data, extra={'user': user, 'record': line_data})
            data.append(line)

        return schema.TableListResult(data=data, total_count=total_count)

    async def retrieve(
            self,
            pk: Any,
            user: auth.UserABC,
            language_context: LanguageContext,
    ) -> schema.RetrieveResult:
        line_data = self._get_data(int(pk))
        line = await self.table_schema.serialize(line_data, extra={'user': user, 'record': line_data})
        return schema.RetrieveResult(data=line)

    async def update(
            self,
            pk: Any,
            data: dict,
            user: auth.UserABC,
            language_context: LanguageContext,
    ) -> schema.UpdateResult:
        logger.info('Updated pk=%s data=%s', pk, data)
        await asyncio.sleep(0.5)
        return schema.UpdateResult(pk=0)

    async def create(
            self,
            data: dict, user:
            auth.UserABC,
            language_context: LanguageContext,
    ) -> schema.CreateResult:
        logger.info('Create data=%s', data)
        return schema.CreateResult(pk=0)



from brilliance_admin import schema, sqlalchemy
from brilliance_admin.translations import TranslateText as _
from example.sections.models import Merchant


class MerchantAdmin(sqlalchemy.SQLAlchemyAdmin):
    model = Merchant
    title = _('merchants')
    icon = 'mdi-card-account-details-outline'

    ordering_fields = [
        'id',
        'user_id',
    ]
    search_fields = [
        'id',
        'title',
    ]

    table_schema = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Merchant,
        readonly_fields=[
            'created_at',
        ],
        fields=[
            'id',
            'user_id',
            'title',
            'description',
            'created_at',
            'terminals',
        ],
        title=schema.StringField(multilined=True, required=True),
        description=schema.StringField(tinymce=True, required=True),
    )
    table_filters = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Merchant,
        fields=[
            'id',
            'user_id',
            'created_at',
            'terminals',
        ],
        created_at=schema.DateTimeField(range=True),
    )



from typing import Any

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import JSONResponse

from brilliance_admin.api.utils import get_category
from brilliance_admin.exceptions import AdminAPIException, APIError
from brilliance_admin.schema import AdminSchema
from brilliance_admin.schema.table.admin_action import ActionData, ActionResult
from brilliance_admin.schema.table.category_table import CategoryTable
from brilliance_admin.schema.table.table_models import CreateResult, ListData, RetrieveResult, TableListResult, UpdateResult
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import get_logger

router = APIRouter(prefix="/table", tags=["Category - Table"])

logger = get_logger()


# pylint: disable=too-many-arguments
@router.post(path='/{group}/{category}/list/')
async def table_list(request: Request, group: str, category: str, list_data: ListData) -> TableListResult:
    schema: AdminSchema = request.app.state.schema

    schema_category, user = await get_category(request, group, category, check_type=CategoryTable)

    language_slug = request.headers.get('Accept-Language')
    language_context: LanguageContext = schema.get_language_context(language_slug)
    context = {'language_context': language_context}

    try:
        result: TableListResult = await schema_category.get_list(list_data, user, language_context)
    except AdminAPIException as e:
        return JSONResponse(e.get_error().model_dump(mode='json', context=context), status_code=e.status_code)

    try:
        return JSONResponse(content=result.model_dump(mode='json', context=context))
    except Exception as e:
        logger.exception('Admin list error: %s; result: %s', e, result)
        raise HTTPException(status_code=500, detail=f"Content error: {e}") from e


@router.post(path='/{group}/{category}/retrieve/{pk}/')
async def table_retrieve(request: Request, group: str, category: str, pk: Any) -> RetrieveResult:
    schema: AdminSchema = request.app.state.schema

    schema_category, user = await get_category(request, group, category, check_type=CategoryTable)
    if not schema_category.has_retrieve:
        raise HTTPException(status_code=404, detail=f"Category {group}.{category} is not allowed for retrive")

    language_slug = request.headers.get('Accept-Language')
    language_context: LanguageContext = schema.get_language_context(language_slug)
    context = {'language_context': language_context}

    try:
        result: RetrieveResult = await schema_category.retrieve(pk, user, language_context)
    except AdminAPIException as e:
        return JSONResponse(e.get_error().model_dump(mode='json', context=context), status_code=e.status_code)

    return JSONResponse(content=result.model_dump(mode='json', context=context))


@router.post(
    path='/{group}/{category}/create/',
    responses={400: {"model": APIError}},
)
async def table_create(request: Request, group: str, category: str) -> CreateResult:
    schema: AdminSchema = request.app.state.schema

    schema_category, user = await get_category(request, group, category, check_type=CategoryTable)
    if not schema_category.has_create:
        raise HTTPException(status_code=404, detail=f"Category {group}.{category} is not allowed for create")

    language_slug = request.headers.get('Accept-Language')
    language_context: LanguageContext = schema.get_language_context(language_slug)
    context = {'language_context': language_context}

    try:
        result: CreateResult = await schema_category.create(await request.json(), user, language_context)
    except AdminAPIException as e:
        return JSONResponse(e.get_error().model_dump(mode='json', context=context), status_code=e.status_code)

    return JSONResponse(content=result.model_dump(mode='json', context=context))


@router.patch(
    path='/{group}/{category}/update/{pk}/',
    responses={400: {"model": APIError}},
)
async def table_update(request: Request, group: str, category: str, pk: Any) -> UpdateResult:
    schema: AdminSchema = request.app.state.schema

    schema_category, user = await get_category(request, group, category, check_type=CategoryTable)
    if not schema_category.has_update:
        raise HTTPException(status_code=404, detail=f"Category {group}.{category} is not allowed for update")

    language_slug = request.headers.get('Accept-Language')
    language_context: LanguageContext = schema.get_language_context(language_slug)
    context = {'language_context': language_context}

    try:
        result: UpdateResult = await schema_category.update(pk, await request.json(), user, language_context)
    except AdminAPIException as e:
        return JSONResponse(e.get_error().model_dump(mode='json', context=context), status_code=e.status_code)

    return JSONResponse(content=result.model_dump(mode='json', context=context))


@router.post(
    path='/{group}/{category}/action/{action}/',
    responses={400: {"model": APIError}},
)
async def table_action(
        request: Request,
        group: str,
        category: str,
        action: str,
        action_data: ActionData,
) -> ActionResult:
    schema: AdminSchema = request.app.state.schema

    schema_category, user = await get_category(request, group, category, check_type=CategoryTable)

    language_slug = request.headers.get('Accept-Language')
    language_context: LanguageContext = schema.get_language_context(language_slug)
    context = {'language_context': language_context}

    try:
        # pylint: disable=protected-access
        result: ActionResult = await schema_category._perform_action(
            request, action, action_data, language_context, user,
        )
    except AdminAPIException as e:
        return JSONResponse(e.get_error().model_dump(mode='json', context=context), status_code=e.status_code)

    return JSONResponse(content=result.model_dump(mode='json', context=context))



from brilliance_admin import schema
from brilliance_admin.auth import UserABC
from brilliance_admin.exceptions import AdminAPIException, APIError
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import get_logger

logger = get_logger()


class SQLAlchemyAdminCreate:
    has_create: bool = True

    async def create(
            self,
            data: dict,
            user: UserABC,
            language_context: LanguageContext,
    ) -> schema.CreateResult:
        if not self.has_create:
            raise AdminAPIException(APIError(message=_('errors.method_not_allowed')), status_code=500)

        # pylint: disable=import-outside-toplevel
        from sqlalchemy.exc import IntegrityError

        try:
            async with self.db_async_session() as session:
                record = await self.table_schema.create(user, data, session)
                pk_value = getattr(record, self.pk_name, None)

        except AdminAPIException as e:
            raise e

        except ConnectionRefusedError as e:
            logger.exception(
                'SQLAlchemy %s create %s db error: %s',
                type(self).__name__, self.table_schema.model.__name__, e,
                extra={'data': data},
            )
            msg = _('errors.connection_refused_error') % {'error': str(e)}
            raise AdminAPIException(
                APIError(message=msg, code='connection_refused_error'),
                status_code=500,
            ) from e

        except IntegrityError as e:
            logger.warning(
                'SQLAlchemy %s create %s db error: %s',
                type(self).__name__, self.table_schema.model.__name__, e,
                extra={'data': data},
            )
            orig = e.orig
            message = orig.args[0] if orig.args else type(orig).__name__
            raise AdminAPIException(
                APIError(message=message, code='db_integrity_error'), status_code=500,
            ) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s create %s db error: %s',
                type(self).__name__, self.table_schema.model.__name__, e,
                extra={'data': data},
            )
            raise AdminAPIException(
                APIError(message=_('errors.db_error_create'), code='db_error_create'), status_code=500,
            ) from e

        logger.info(
            '%s model %s #%s created by %s',
            type(self).__name__, self.table_schema.model.__name__, pk_value, user.username,
            extra={'data': data},
        )
        return schema.CreateResult(pk=pk_value)



from typing import Any

from brilliance_admin.integrations.sqlalchemy.autocomplete import SQLAlchemyAdminAutocompleteMixin
from brilliance_admin.integrations.sqlalchemy.fields_schema import SQLAlchemyFieldsSchema
from brilliance_admin.schema.table.category_table import CategoryTable
from brilliance_admin.translations import TranslateText as _

EXCEPTION_REL_NAME = '''
Model "{model_name}" doesn\'t contain rel_name:"{rel_name}" for field "{slug}"
Model fields = {model_attrs}
'''


class SQLAlchemyAdminBase(SQLAlchemyAdminAutocompleteMixin, CategoryTable):
    model: Any
    slug = None
    ordering_fields = []

    search_fields = []

    table_schema: SQLAlchemyFieldsSchema

    db_async_session: Any = None

    def __init__(
            self,
            *args,
            model=None,
            table_schema=None,
            db_async_session=None,
            ordering_fields=None,
            default_ordering=None,
            search_fields=None,
            **kwargs,
    ):
        if model:
            self.model = model

        if search_fields:
            self.search_fields = search_fields

        if self.search_fields:
            self.search_enabled = True
            self.search_help = _('sqlalchemy_search_help') % {'fields': ', '.join(self.search_fields)}

        if default_ordering:
            self.default_ordering = default_ordering

        if ordering_fields:
            self.ordering_fields = ordering_fields

        self.validate_fields()

        if table_schema:
            self.table_schema = table_schema

        if not self.table_schema:
            self.table_schema = SQLAlchemyFieldsSchema(model=self.model)

        if not issubclass(type(self.table_schema), SQLAlchemyFieldsSchema):
            msg = f'{type(self).__name__}.table_schema {self.table_schema} must be subclass of SQLAlchemyFieldsSchema'
            raise AttributeError(msg)

        if not self.model:
            msg = f'{type(self).__name__}.model is required for SQLAlchemy'
            raise AttributeError(msg)

        if not self.slug:
            self.slug = self.model.__name__.lower()

        if db_async_session:
            self.db_async_session = db_async_session

        if not self.db_async_session:
            msg = f'{type(self).__name__}.db_async_session is required for SQLAlchemy'
            raise AttributeError(msg)

        # pylint: disable=import-outside-toplevel
        from sqlalchemy import inspect
        from sqlalchemy.sql.schema import Column

        for attr in inspect(self.model).mapper.column_attrs:
            col: Column = attr.columns[0]
            if col.primary_key and not self.pk_name:
                self.pk_name = attr.key
                break

        if not self.default_ordering and self.pk_name:
            self.default_ordering = f'-{self.pk_name}'

        super().__init__(*args, **kwargs)

    def validate_fields(self):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy.orm import InstrumentedAttribute

        if self.search_fields:
            for field in self.search_fields:
                column = getattr(self.model, field, None)
                if not isinstance(column, InstrumentedAttribute):
                    raise AttributeError(
                        f'{type(self).__name__}: search field "{field}" not found in model {self.model.__name__}'
                    )

        if self.ordering_fields:
            for field in self.ordering_fields:
                column = getattr(self.model, field, None)
                if not isinstance(column, InstrumentedAttribute):
                    raise AttributeError(
                        f'{type(self).__name__}: ordering field "{field}" not found in model {self.model.__name__}'
                    )

    def get_queryset(self):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import select
        from sqlalchemy.orm import selectinload

        stmt = select(self.model).options(selectinload('*'))

        # Eager-load related fields
        for slug, field in self.table_schema.get_fields().items():

            # pylint: disable=protected-access
            if field._type == "related":

                if not hasattr(self.model, field.rel_name):
                    # pylint: disable=import-outside-toplevel
                    from sqlalchemy import inspect
                    model_attrs = [attr.key for attr in inspect(self.model).mapper.attrs]

                    msg = EXCEPTION_REL_NAME.format(
                        slug=slug,
                        model_name=self.model.__name__,
                        rel_name=field.rel_name,
                        model_attrs=model_attrs,
                    )
                    raise AttributeError(msg)

                stmt = stmt.options(selectinload(getattr(self.model, field.rel_name)))

        return stmt



from brilliance_admin import sqlalchemy
from brilliance_admin.translations import TranslateText as _
from example.sections.models import Currency


class CurrencyAdmin(sqlalchemy.SQLAlchemyAdmin):
    model = Currency
    title = _('currencies')
    icon = 'mdi-currency-usd'

    table_schema = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Currency,
        readonly_fields=['terminals'],
    )
    table_filters = sqlalchemy.SQLAlchemyFieldsSchema(
        model=Currency,
        fields=[
            'id',
            'terminals',
        ],
    )



import logging
import logging.config

import structlog
from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from structlog.dev import RichTracebackFormatter

from brilliance_admin import schema
from brilliance_admin.auth import AdminAuthentication, AuthData, AuthResult, UserABC, UserResult
from brilliance_admin.exceptions import AdminAPIException, APIError
from brilliance_admin.translations import LanguageManager
from brilliance_admin.translations import TranslateText as _
from example.sections.currency import CurrencyAdmin
from example.sections.graphs import GraphsExample
from example.sections.merchant import MerchantAdmin
from example.sections.payments import PaymentsAdmin
from example.sections.terminal import TerminalAdmin
from example.sections.users import UserAdmin
from example.sqlite import async_sessionmaker_, lifespan

structlog.configure(
    processors=[
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.stdlib.ExtraAdder(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S", utc=False),
        structlog.dev.ConsoleRenderer(colors=True, sort_keys=True)
    ],
    wrapper_class=structlog.make_filtering_bound_logger(logging.NOTSET),
    context_class=dict,
    logger_factory=structlog.PrintLoggerFactory(),
    cache_logger_on_first_use=False
)

cr = structlog.dev.ConsoleRenderer.get_active()
cr.exception_formatter = RichTracebackFormatter(
    max_frames=1,
    show_locals=False,
    suppress=["uvicorn", "fastapi", "starlette"],
)


LOGIN_GREETINGS_MESSAGE = '''
<div class="text-h6 mb-1">
  Demo mode
</div>
<div class="text-caption">
  Login: admin<br>
  Password: admin
</div>
'''


class FakeAdminAuthentication(AdminAuthentication):
    async def login(self, data: AuthData) -> AuthResult:
        if data.username != 'admin' or data.password != 'admin':
            raise AdminAPIException(APIError(code='user_not_found'), status_code=401)

        return AuthResult(token='test', user=UserResult(username='test_admin'))

    async def authenticate(self, headers: dict) -> UserABC:
        return UserABC(username='test_admin')


admin_schema = schema.AdminSchema(
    title=_('admin_title'),
    description=_('admin_description'),
    login_greetings_message=_('login_greetings_message'),

    logo_image='/static/logo-outline.png',
    favicon_image='/static/favicon.jpg',

    auth=FakeAdminAuthentication(),
    language_manager=LanguageManager(
        locales_dir='example/locales',
        languages={
            'ru': 'Russian',
            'en': 'English',
            'test': 'Test',
        },
    ),

    groups=[
        schema.Group(
            slug='payments',
            title=_('payments'),
            icon='mdi-cash-multiple',
            categories=[
                PaymentsAdmin(),
            ]
        ),
        schema.Group(
            slug='users',
            title=_('users'),
            icon='mdi-account',
            categories=[
                UserAdmin(db_async_session=async_sessionmaker_),
            ]
        ),
        schema.Group(
            slug='merchants',
            title=_('merchants'),
            icon='mdi-folder-account-outline',
            categories=[
                MerchantAdmin(db_async_session=async_sessionmaker_),
                TerminalAdmin(db_async_session=async_sessionmaker_),
            ]
        ),
        schema.Group(
            slug='currencies',
            title=_('currencies'),
            icon='mdi-cash-multiple',
            categories=[
                CurrencyAdmin(db_async_session=async_sessionmaker_),
            ]
        ),
        schema.Group(
            slug='statistics',
            title=_('statistics'),
            icon='mdi-finance',
            categories=[
                GraphsExample(),
            ]
        ),
    ],
)


app = FastAPI(debug=True, lifespan=lifespan)
app.mount('/static', StaticFiles(directory='example/static'), name='static')

admin_app = admin_schema.generate_app(
    include_scalar=True, include_docs=True, include_redoc=True, debug=True,
)
app.mount('/admin', admin_app)


@app.get('/')
async def root():
    return RedirectResponse(url='/admin/')



import abc
from typing import Any, ClassVar, Dict, List

from pydantic import Field
from pydantic.dataclasses import dataclass
from pydantic_core import core_schema

from brilliance_admin.auth import UserABC
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import DataclassBase, SupportsStr


# pylint: disable=too-many-instance-attributes
@dataclass
class FieldSchemaData(DataclassBase):
    type: str = None

    label: str | None = None
    help_text: str | None = None

    # Table header parameters
    header: dict = Field(default_factory=dict)

    read_only: bool = False
    default: Any | None = None
    required: bool = False

    max_length: int | None = None
    min_length: int | None = None

    choices: List[dict] | None = None

    variant: str | None = None
    size: str | None = None

    preview_max_height: int | None = None
    preview_max_width: int | None = None

    # StringField
    multilined: bool | None = None
    ckeditor: bool | None = None
    tinymce: bool | None = None

    # ArrayField
    array_type: str | None = None

    # SQLAlchemyRelatedField
    many: bool | None = None
    rel_name: str | None = None
    dual_list: bool | None = None

    # IntegerField
    inputmode: str | None = None
    precision: int | None = None
    scale: int | None = None

    # DateTimeField
    range: bool | None = None
    include_date: bool | None = None
    include_time: bool | None = None


@dataclass
class FieldsSchemaData(DataclassBase):
    fields: Dict[str, dict] = Field(default_factory=dict)
    list_display: List[str] = Field(default_factory=list)


# pylint: disable=too-many-instance-attributes
@dataclass
class TableInfoSchemaData(DataclassBase):
    table_schema: FieldsSchemaData

    search_enabled: bool = Field(default=False)
    search_help: str | None = Field(default=None)

    pk_name: str | None = Field(default=None)
    can_retrieve: bool = Field(default=False)

    can_create: bool = Field(default=False)
    can_update: bool = Field(default=False)

    table_filters: FieldsSchemaData | None = Field(default=None)

    ordering_fields: List[str] = Field(default_factory=list)
    default_ordering: str | None = None

    actions: Dict[str, dict] | None = Field(default_factory=dict)

    def __repr__(self):
        return f'<TableInfoSchemaData id={id(self)}>'


@dataclass
class GraphInfoSchemaData(DataclassBase):
    search_enabled: bool
    search_help: str | None

    table_filters: FieldsSchemaData | None = None


@dataclass
class CategorySchemaData(DataclassBase):
    title: str | None
    description: str | None
    icon: str | None
    type: str

    table_info: TableInfoSchemaData | None = None
    graph_info: GraphInfoSchemaData | None = None

    def __repr__(self):
        return f'<CategorySchemaData type={self.type} "{self.title}">'


class Category(abc.ABC):
    slug: ClassVar[str]
    title: ClassVar[SupportsStr | None] = None
    description: ClassVar[SupportsStr | None] = None

    # https://pictogrammers.com/library/mdi/
    icon: ClassVar[str | None] = None

    _type_slug: ClassVar[str]

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> CategorySchemaData:
        return CategorySchemaData(
            title=language_context.get_text(self.title) or self.slug,
            description=language_context.get_text(self.description),
            icon=self.icon,
            type=self._type_slug,
        )

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: Any) -> core_schema.CoreSchema:
        def validate(v: Any) -> "Category":
            if isinstance(v, cls):
                return v
            raise TypeError(f"Expected {cls.__name__} instance")

        return core_schema.no_info_plain_validator_function(
            validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: repr(v),
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )



from fastapi import APIRouter

from .views.schema import router as schema_router
from .views.table import router as schema_table
from .views.auth import router as schema_auth
from .views.autocomplete import router as schema_autocomplete
from .views.graphs import router as schema_graphs
from .views.settings import router as schema_settings
from .views.index import router as schema_index

brilliance_admin_router = APIRouter()
brilliance_admin_router.include_router(schema_router)
brilliance_admin_router.include_router(schema_table)
brilliance_admin_router.include_router(schema_auth)
brilliance_admin_router.include_router(schema_autocomplete)
brilliance_admin_router.include_router(schema_graphs)
brilliance_admin_router.include_router(schema_settings)
brilliance_admin_router.include_router(schema_index)
